"""
REST API — mounted under /api/v1 by server.py.

All field names use camelCase in JSON to match the TypeScript frontend.
"""
from __future__ import annotations

import mimetypes
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Optional

from fastapi import APIRouter, HTTPException, Query, UploadFile, File
from fastapi.responses import FileResponse
from pydantic import BaseModel, ConfigDict

from ..storage.base import Company, CompanyAddress, PasswordEntry
from ..storage.sqlite import SQLiteStorage, project_dir

router = APIRouter(prefix="/api/v1")

_sessions: dict[str, SQLiteStorage] = {}

ALLOWED_IMAGE_TYPES = {"image/png", "image/svg+xml", "image/webp", "image/jpeg"}
MAX_ICON_BYTES = 2 * 1024 * 1024


def set_storage(storage: SQLiteStorage) -> None:
    _sessions["default"] = storage


def _require(project: str) -> SQLiteStorage:
    s = _sessions.get(project)
    if s is None:
        raise HTTPException(
            status_code=401,
            detail=f"Project '{project}' is locked. POST /api/v1/projects/{project}/unlock first.",
        )
    return s


# ── Pydantic schemas ──────────────────────────────────────────────────────────

class AddressSchema(BaseModel):
    street:      Optional[str] = None
    city:        Optional[str] = None
    state:       Optional[str] = None
    postcode:    Optional[str] = None
    country:     str           = ""
    countryCode: str           = ""


class CompanyIn(BaseModel):
    name:    str
    icon:    Optional[dict] = None          # {type, value}
    address: Optional[AddressSchema] = None
    revenue: Optional[float] = None


class CompanyOut(CompanyIn):
    id: int
    model_config = ConfigDict(from_attributes=True)


class EntryIn(BaseModel):
    title:          str
    username:       Optional[str]       = None
    email:          Optional[str]       = None
    password:       Optional[str]       = None
    url:            Optional[str]       = None
    notes:          Optional[str]       = None
    category:       str                 = "Other"
    tags:           list[str]           = []
    login_methods:  list[str]           = []
    company_id:     Optional[int]       = None
    user_created_at: Optional[str]      = None   # ISO string


class EntryOut(EntryIn):
    id:         int
    created_at: str
    updated_at: str
    model_config = ConfigDict(from_attributes=True)


class UnlockIn(BaseModel):
    password: str


class ProjectIn(BaseModel):
    name:     str
    password: str


class IconOut(BaseModel):
    filename: str
    url:      str


# ── helpers ────────────────────────────────────────────────────────────────────

def _parse_dt(s: Optional[str]) -> Optional[datetime]:
    if not s:
        return None
    try:
        return datetime.fromisoformat(s)
    except ValueError:
        return None


def _entry_in_to_obj(body: EntryIn) -> PasswordEntry:
    return PasswordEntry(
        id              = None,
        title           = body.title,
        username        = body.username,
        email           = body.email,
        password        = body.password,
        url             = body.url,
        notes           = body.notes,
        category        = body.category or "Other",
        tags            = body.tags or [],
        login_methods   = body.login_methods or [],
        company_id      = body.company_id,
        user_created_at = _parse_dt(body.user_created_at),
    )


def _company_in_to_obj(body: CompanyIn, company_id: Optional[int] = None) -> Company:
    addr = None
    if body.address and body.address.country:
        addr = CompanyAddress(
            country      = body.address.country,
            country_code = body.address.countryCode,
            street       = body.address.street,
            city         = body.address.city,
            state        = body.address.state,
            postcode     = body.address.postcode,
        )
    return Company(
        id      = company_id,
        name    = body.name,
        icon    = body.icon,
        address = addr,
        revenue = body.revenue,
    )


# ── projects ──────────────────────────────────────────────────────────────────

@router.get("/projects", response_model=list[str])
def list_projects():
    root = Path.home() / ".sspwd"
    if not root.exists():
        return []
    return sorted(
        d.name for d in root.iterdir()
        if d.is_dir() and (d / "vault.db").exists()
    )


@router.get("/projects/unlocked", response_model=list[str])
def list_unlocked():
    return list(_sessions.keys())


@router.post("/projects/{name}/unlock", status_code=200)
def unlock_project(name: str, body: UnlockIn):
    vault = project_dir(name)
    if not (vault / "vault.db").exists():
        raise HTTPException(status_code=404, detail=f"Project '{name}' does not exist.")
    try:
        storage = SQLiteStorage(master_password=body.password, project=name)
        _sessions[name] = storage
    except Exception:
        raise HTTPException(status_code=401, detail="Wrong master password.")
    return {"project": name, "status": "unlocked"}


@router.post("/projects", status_code=201)
def create_project(body: ProjectIn):
    name = body.name.strip()
    if not name or "/" in name or "\\" in name:
        raise HTTPException(status_code=400, detail="Invalid project name.")
    vault = project_dir(name)
    if (vault / "vault.db").exists():
        raise HTTPException(status_code=409, detail=f"Project '{name}' already exists.")
    storage = SQLiteStorage(master_password=body.password, project=name)
    _sessions[name] = storage
    return {"project": name, "status": "created"}


@router.delete("/projects/{name}/lock", status_code=200)
def lock_project(name: str):
    _sessions.pop(name, None)
    return {"project": name, "status": "locked"}


# ── entries ────────────────────────────────────────────────────────────────────

@router.get("/entries", response_model=list[EntryOut])
def list_entries(project: str = Query(...), search: Optional[str] = Query(default=None)):
    return [e.to_dict() for e in _require(project).list(search=search)]


@router.post("/entries", response_model=EntryOut, status_code=201)
def create_entry(body: EntryIn, project: str = Query(...)):
    return _require(project).add(_entry_in_to_obj(body)).to_dict()


@router.get("/entries/{entry_id}", response_model=EntryOut)
def get_entry(entry_id: int, project: str = Query(...)):
    entry = _require(project).get(entry_id)
    if entry is None:
        raise HTTPException(status_code=404, detail="Entry not found.")
    return entry.to_dict()


@router.put("/entries/{entry_id}", response_model=EntryOut)
def update_entry(entry_id: int, body: EntryIn, project: str = Query(...)):
    obj    = _entry_in_to_obj(body)
    obj.id = entry_id
    try:
        return _require(project).update(obj).to_dict()
    except KeyError:
        raise HTTPException(status_code=404, detail="Entry not found.")


@router.delete("/entries/{entry_id}", status_code=204)
def delete_entry(entry_id: int, project: str = Query(...)):
    try:
        _require(project).delete(entry_id)
    except KeyError:
        raise HTTPException(status_code=404, detail="Entry not found.")


# ── companies ──────────────────────────────────────────────────────────────────

@router.get("/companies", response_model=list[CompanyOut])
def list_companies(project: str = Query(...)):
    return [c.to_dict() for c in _require(project).list_companies()]


@router.post("/companies", response_model=CompanyOut, status_code=201)
def create_company(body: CompanyIn, project: str = Query(...)):
    return _require(project).add_company(_company_in_to_obj(body)).to_dict()


@router.get("/companies/{company_id}", response_model=CompanyOut)
def get_company(company_id: int, project: str = Query(...)):
    c = _require(project).get_company(company_id)
    if c is None:
        raise HTTPException(status_code=404, detail="Company not found.")
    return c.to_dict()


@router.put("/companies/{company_id}", response_model=CompanyOut)
def update_company(company_id: int, body: CompanyIn, project: str = Query(...)):
    try:
        return _require(project).update_company(
            _company_in_to_obj(body, company_id)
        ).to_dict()
    except KeyError:
        raise HTTPException(status_code=404, detail="Company not found.")


@router.delete("/companies/{company_id}", status_code=204)
def delete_company(company_id: int, project: str = Query(...)):
    try:
        _require(project).delete_company(company_id)
    except KeyError:
        raise HTTPException(status_code=404, detail="Company not found.")


# ── icons ──────────────────────────────────────────────────────────────────────

@router.post("/icons", response_model=IconOut, status_code=201)
async def upload_icon(file: UploadFile = File(...), project: str = Query(...)):
    storage = _require(project)
    ct = file.content_type or ""
    if ct not in ALLOWED_IMAGE_TYPES:
        raise HTTPException(status_code=415, detail=f"Unsupported type '{ct}'.")
    data = await file.read()
    if len(data) > MAX_ICON_BYTES:
        raise HTTPException(status_code=413, detail="Icon exceeds 2 MB.")
    ext = mimetypes.guess_extension(ct) or ".png"
    if ext in (".jpe", ".jfif"):
        ext = ".jpg"
    filename = f"{uuid.uuid4().hex}{ext}"
    storage.save_icon(filename, data)
    return {"filename": filename, "url": f"/api/v1/icons/{filename}?project={project}"}


@router.get("/icons/{filename}")
def serve_icon(filename: str, project: str = Query(...)):
    storage = _require(project)
    p = storage.icons_dir / filename
    if not p.exists():
        raise HTTPException(status_code=404, detail="Icon not found.")
    try:
        p.relative_to(storage.icons_dir)
    except ValueError:
        raise HTTPException(status_code=400, detail="Invalid filename.")
    return FileResponse(p)


@router.get("/icons", response_model=list[IconOut])
def list_icons(project: str = Query(...)):
    s = _require(project)
    return [
        {"filename": n, "url": f"/api/v1/icons/{n}?project={project}"}
        for n in s.list_icons()
    ]


# ── health ─────────────────────────────────────────────────────────────────────

@router.get("/health")
def health():
    return {"status": "ok"}


"""
CLI for sspwd.

Commands
--------
    sspwd serve   [--project NAME]   Start the web UI
    sspwd add     [--project NAME]   Add entry interactively
    sspwd list    [--project NAME]   List entries
    sspwd get     [--project NAME]   Show one entry
    sspwd delete  [--project NAME]   Delete an entry
    sspwd version                    Print version
    sspwd projects                   List existing projects

Projects are stored as separate vaults under ~/.sspwd/{project}/vault.db.
The default project is named "default".
"""

import sys
from pathlib import Path
from typing import Optional

import click

from .__version__ import __version__
from .storage.sqlite import SQLiteStorage, project_dir, _DEFAULT_PROJECT
from .storage.base import PasswordEntry

# ---------------------------------------------------------------------------
# Shared option decorator
# ---------------------------------------------------------------------------

_project_option = click.option(
    "--project", "-p",
    default=_DEFAULT_PROJECT,
    show_default=True,
    help="Project / workspace name (subdirectory of ~/.sspwd/).",
)


def _get_storage(master: str, project: str, vault_dir: Optional[Path] = None) -> SQLiteStorage:
    return SQLiteStorage(master_password=master, project=project, vault_dir=vault_dir)


def _prompt_master() -> str:
    return click.prompt("Master password", hide_input=True)


# ---------------------------------------------------------------------------
# Root group
# ---------------------------------------------------------------------------


@click.group()
@click.version_option(__version__, prog_name="sspwd")
def cli() -> None:
    """sspwd — a local, encrypted password manager."""


# ---------------------------------------------------------------------------
# Commands
# ---------------------------------------------------------------------------


@cli.command()
@click.option("--host", default="127.0.0.1", show_default=True)
@click.option("--port", default=7523, show_default=True)
@click.option("--no-browser", is_flag=True, default=False)
def serve(host: str, port: int, no_browser: bool) -> None:
    """Start the web UI. Projects are unlocked on demand in the browser."""
    from .ui.server import UIServer
    server = UIServer(host=host, port=port, open_browser=not no_browser)
    server.start(block=True)


@cli.command("add")
@_project_option
@click.option("--vault-dir", default=None, type=click.Path())
def add_entry(project: str, vault_dir: Optional[str]) -> None:
    """Add a new password entry interactively."""
    master  = _prompt_master()
    storage = _get_storage(master, project, Path(vault_dir) if vault_dir else None)

    title    = click.prompt("Title")
    category = click.prompt("Category")
    username = click.prompt("Username")
    password = click.prompt("Password", hide_input=True, confirmation_prompt=True)
    url      = click.prompt("URL (optional)", default="")
    notes    = click.prompt("Notes (optional)", default="")

    entry = PasswordEntry(
        id=None, title=title, username=username, password=password,
        url=url or None, notes=notes or None,
    )
    created = storage.add(entry)
    click.echo(click.style(f"✓ Saved '{title}' (id={created.id}) in project '{project}'", fg="green"))


@cli.command("list")
@_project_option
@click.option("--search", "-s", default=None, help="Filter by title / username / URL.")
@click.option("--vault-dir", default=None, type=click.Path())
def list_entries(project: str, search: Optional[str], vault_dir: Optional[str]) -> None:
    """List all stored entries for a project."""
    master  = _prompt_master()
    storage = _get_storage(master, project, Path(vault_dir) if vault_dir else None)

    entries = storage.list(search=search)
    if not entries:
        click.echo("No entries found.")
        return

    click.echo(f"\n  Project: {project}\n")
    click.echo(f"{'ID':>4}  {'Title':<30}  {'Username':<25}  URL")
    click.echo("-" * 80)
    for e in entries:
        click.echo(f"{e.id:>4}  {e.title:<30}  {e.username:<25}  {e.url or ''}")


@cli.command("get")
@_project_option
@click.argument("entry_id", type=int)
@click.option("--vault-dir", default=None, type=click.Path())
def get_entry(project: str, entry_id: int, vault_dir: Optional[str]) -> None:
    """Show a single entry (reveals password)."""
    master  = _prompt_master()
    storage = _get_storage(master, project, Path(vault_dir) if vault_dir else None)
    entry   = storage.get(entry_id)

    if entry is None:
        click.echo(click.style(f"Entry {entry_id} not found.", fg="red"), err=True)
        sys.exit(1)

    click.echo(f"\nTitle    : {entry.title}")
    click.echo(f"Username : {entry.username}")
    click.echo(f"Password : {entry.password}")
    click.echo(f"URL      : {entry.url or '-'}")
    click.echo(f"Notes    : {entry.notes or '-'}")
    click.echo(f"Updated  : {entry.updated_at}")


@cli.command("delete")
@_project_option
@click.argument("entry_id", type=int)
@click.option("--yes", "-y", is_flag=True, default=False)
@click.option("--vault-dir", default=None, type=click.Path())
def delete_entry(project: str, entry_id: int, yes: bool, vault_dir: Optional[str]) -> None:
    """Delete an entry by ID."""
    master  = _prompt_master()
    storage = _get_storage(master, project, Path(vault_dir) if vault_dir else None)

    if not yes:
        click.confirm(f"Delete entry {entry_id} from project '{project}'?", abort=True)

    try:
        storage.delete(entry_id)
        click.echo(click.style(f"✓ Deleted entry {entry_id}.", fg="green"))
    except KeyError:
        click.echo(click.style(f"Entry {entry_id} not found.", fg="red"), err=True)
        sys.exit(1)


@cli.command("projects")
def list_projects() -> None:
    """List all existing projects under ~/.sspwd/."""
    root = Path.home() / ".sspwd"
    if not root.exists():
        click.echo("No projects found (vault directory does not exist yet).")
        return

    projects = [d.name for d in sorted(root.iterdir()) if d.is_dir() and (d / "vault.db").exists()]
    if not projects:
        click.echo("No projects found.")
        return

    click.echo(f"\n  Vault: {root}\n")
    for name in projects:
        size = (root / name / "vault.db").stat().st_size
        click.echo(f"  {'●'} {name:<30} ({size:,} bytes)")


@cli.command("version")
def version() -> None:
    """Print version."""
    click.echo(f"sspwd {__version__}")


def main() -> None:
    cli()


"""Tests for the SQLite storage backend — targets 100% storage coverage."""

import pytest
from pathlib import Path
from datetime import datetime

from sspwd.storage.sqlite import SQLiteStorage
from sspwd.storage.base import BaseStorage, Company, CompanyAddress, PasswordEntry


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def storage(tmp_path: Path) -> SQLiteStorage:
    return SQLiteStorage(master_password="test-master", vault_dir=tmp_path)


def _entry(**kwargs) -> PasswordEntry:
    defaults = dict(id=None, title="GitHub", username="alice", password="s3cr3t")
    defaults.update(kwargs)
    return PasswordEntry(**defaults)


def _company(**kwargs) -> Company:
    defaults = dict(id=None, name="Acme Corp")
    defaults.update(kwargs)
    return Company(**defaults)


# ---------------------------------------------------------------------------
# CompanyAddress dataclass
# ---------------------------------------------------------------------------


class TestCompanyAddress:
    def test_to_dict_full(self) -> None:
        a = CompanyAddress(
            country="United States", country_code="us",
            street="1 Main St", city="Springfield", state="IL", postcode="62701",
        )
        d = a.to_dict()
        assert d["country"] == "United States"
        assert d["countryCode"] == "us"
        assert d["street"] == "1 Main St"
        assert d["city"] == "Springfield"
        assert d["state"] == "IL"
        assert d["postcode"] == "62701"

    def test_to_dict_minimal(self) -> None:
        a = CompanyAddress(country="Germany", country_code="de")
        d = a.to_dict()
        assert d["country"] == "Germany"
        assert d["street"] is None

    def test_from_dict_round_trip(self) -> None:
        original = CompanyAddress(
            country="France", country_code="fr",
            street="1 Rue de Rivoli", city="Paris", postcode="75001",
        )
        restored = CompanyAddress.from_dict(original.to_dict())
        assert restored.country == original.country
        assert restored.country_code == original.country_code
        assert restored.street == original.street
        assert restored.city == original.city

    def test_from_dict_missing_keys(self) -> None:
        a = CompanyAddress.from_dict({"country": "Japan", "countryCode": "jp"})
        assert a.street is None
        assert a.state is None


# ---------------------------------------------------------------------------
# Company dataclass
# ---------------------------------------------------------------------------


class TestCompanyDataclass:
    def test_to_dict_with_address(self) -> None:
        addr = CompanyAddress(country="US", country_code="us", city="NYC")
        c = Company(id=1, name="Acme", icon={"type": "letter", "value": "A"},
                    address=addr, revenue=1_000_000.0)
        d = c.to_dict()
        assert d["name"] == "Acme"
        assert d["address"]["city"] == "NYC"
        assert d["revenue"] == 1_000_000.0
        assert d["icon"]["type"] == "letter"

    def test_to_dict_no_address(self) -> None:
        c = Company(id=2, name="NoAddr")
        d = c.to_dict()
        assert d["address"] is None
        assert d["revenue"] is None

    def test_from_dict_round_trip(self) -> None:
        addr = CompanyAddress(country="Canada", country_code="ca", city="Toronto")
        c = Company(id=5, name="CanadaCo", icon={"type": "url", "value": "https://x.com/logo.svg"},
                    address=addr, revenue=5_000_000.0)
        restored = Company.from_dict(c.to_dict())
        assert restored.name == "CanadaCo"
        assert restored.revenue == 5_000_000.0
        assert restored.address is not None
        assert restored.address.country == "Canada"

    def test_from_dict_no_address(self) -> None:
        c = Company.from_dict({"id": 1, "name": "Plain"})
        assert c.address is None
        assert c.icon is None


# ---------------------------------------------------------------------------
# BaseStorage abstract NotImplementedError methods
# ---------------------------------------------------------------------------


class TestBaseStorageDefaults:
    """Verify that the default company methods raise NotImplementedError."""

    def test_add_company_not_implemented(self, tmp_path: Path) -> None:
        s = SQLiteStorage.__new__(SQLiteStorage)
        with pytest.raises(NotImplementedError):
            BaseStorage.add_company(s, _company())

    def test_get_company_not_implemented(self, tmp_path: Path) -> None:
        s = SQLiteStorage.__new__(SQLiteStorage)
        with pytest.raises(NotImplementedError):
            BaseStorage.get_company(s, 1)

    def test_list_companies_not_implemented(self, tmp_path: Path) -> None:
        s = SQLiteStorage.__new__(SQLiteStorage)
        with pytest.raises(NotImplementedError):
            BaseStorage.list_companies(s)

    def test_update_company_not_implemented(self, tmp_path: Path) -> None:
        s = SQLiteStorage.__new__(SQLiteStorage)
        with pytest.raises(NotImplementedError):
            BaseStorage.update_company(s, _company(id=1))

    def test_delete_company_not_implemented(self, tmp_path: Path) -> None:
        s = SQLiteStorage.__new__(SQLiteStorage)
        with pytest.raises(NotImplementedError):
            BaseStorage.delete_company(s, 1)


# ---------------------------------------------------------------------------
# Initialisation
# ---------------------------------------------------------------------------


class TestInit:
    def test_creates_vault_dir(self, tmp_path: Path) -> None:
        vault = tmp_path / "nested" / "vault"
        SQLiteStorage(master_password="pw", vault_dir=vault)
        assert vault.is_dir()

    def test_creates_db_file(self, tmp_path: Path) -> None:
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        assert (tmp_path / "vault.db").exists()

    def test_creates_salt_file(self, tmp_path: Path) -> None:
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        assert (tmp_path / "salt.bin").exists()

    def test_salt_reused_on_second_init(self, tmp_path: Path) -> None:
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        salt1 = (tmp_path / "salt.bin").read_bytes()
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        salt2 = (tmp_path / "salt.bin").read_bytes()
        assert salt1 == salt2

    def test_default_vault_dir_is_home_sspwd(self, monkeypatch, tmp_path) -> None:
        monkeypatch.setattr(Path, "home", lambda: tmp_path)
        s = SQLiteStorage(master_password="pw")
        assert s._vault_dir == tmp_path / ".sspwd" / "default"

    def test_vault_dir_property(self, tmp_path: Path) -> None:
        s = SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        assert s.vault_dir == tmp_path

    def test_migration_adds_missing_entry_columns(self, tmp_path: Path) -> None:
        """Migration runs ALTER TABLE for any column absent in an older vault."""
        import sqlite3
        # Create a minimal old-style DB with only legacy columns
        db = tmp_path / "vault.db"
        conn = sqlite3.connect(db)
        conn.execute(
            "CREATE TABLE entries (id INTEGER PRIMARY KEY, title TEXT NOT NULL,"
            " username TEXT, password TEXT, url TEXT, notes TEXT,"
            " created_at TEXT NOT NULL, updated_at TEXT NOT NULL)"
        )
        conn.execute(
            "CREATE TABLE companies (id INTEGER PRIMARY KEY, name TEXT NOT NULL)"
        )
        conn.commit()
        conn.close()
        # Initialising against this old DB should apply the migration
        (tmp_path / "salt.bin").write_bytes(b"\x00" * 32)
        s = SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        cols = {row[1] for row in s._connect().execute("PRAGMA table_info(entries)")}
        assert "email" in cols
        assert "category" in cols
        assert "tags" in cols


# ---------------------------------------------------------------------------
# Add
# ---------------------------------------------------------------------------


class TestAdd:
    def test_returns_entry_with_id(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry())
        assert entry.id is not None and entry.id > 0

    def test_increments_ids(self, storage: SQLiteStorage) -> None:
        e1 = storage.add(_entry(title="A"))
        e2 = storage.add(_entry(title="B"))
        assert e1.id != e2.id

    def test_password_round_trip(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(password="super-secret"))
        assert storage.get(entry.id).password == "super-secret"

    def test_email_round_trip(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(email="alice@example.com"))
        assert storage.get(entry.id).email == "alice@example.com"

    def test_notes_round_trip(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(notes="personal account"))
        assert storage.get(entry.id).notes == "personal account"

    def test_category_stored(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(category="Finance"))
        assert storage.get(entry.id).category == "Finance"

    def test_tags_round_trip(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(tags=["work", "critical"]))
        assert storage.get(entry.id).tags == ["work", "critical"]

    def test_login_methods_round_trip(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(login_methods=["Google", "Email / Password"]))
        assert storage.get(entry.id).login_methods == ["Google", "Email / Password"]

    def test_optional_fields_none(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(url=None, notes=None, email=None))
        fetched = storage.get(entry.id)
        assert fetched.url is None
        assert fetched.notes is None
        assert fetched.email is None

    def test_timestamps_set(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry())
        assert isinstance(entry.created_at, datetime)
        assert isinstance(entry.updated_at, datetime)

    def test_title_only_entry(self, storage: SQLiteStorage) -> None:
        """All fields except title are optional."""
        e = storage.add(PasswordEntry(id=None, title="Bare"))
        fetched = storage.get(e.id)
        assert fetched.title == "Bare"
        assert fetched.username is None
        assert fetched.password is None


# ---------------------------------------------------------------------------
# Get
# ---------------------------------------------------------------------------


class TestGet:
    def test_returns_none_for_missing(self, storage: SQLiteStorage) -> None:
        assert storage.get(99999) is None

    def test_returns_correct_entry(self, storage: SQLiteStorage) -> None:
        added = storage.add(_entry(title="Notion", username="bob"))
        fetched = storage.get(added.id)
        assert fetched.title == "Notion"
        assert fetched.username == "bob"

    def test_fields_complete(self, storage: SQLiteStorage) -> None:
        added = storage.add(_entry(url="https://example.com", notes="note"))
        fetched = storage.get(added.id)
        assert fetched.url == "https://example.com"
        assert fetched.notes == "note"


# ---------------------------------------------------------------------------
# List
# ---------------------------------------------------------------------------


class TestList:
    def test_empty(self, storage: SQLiteStorage) -> None:
        assert storage.list() == []

    def test_returns_all(self, storage: SQLiteStorage) -> None:
        for t in ("A", "B", "C"):
            storage.add(_entry(title=t))
        assert len(storage.list()) == 3

    def test_ordered_by_title(self, storage: SQLiteStorage) -> None:
        for t in ("Zebra", "Apple", "Mango"):
            storage.add(_entry(title=t))
        titles = [e.title for e in storage.list()]
        assert titles == sorted(titles)

    def test_search_by_title(self, storage: SQLiteStorage) -> None:
        storage.add(_entry(title="GitHub"))
        storage.add(_entry(title="GitLab"))
        storage.add(_entry(title="AWS"))
        assert len(storage.list(search="Git")) == 2

    def test_search_by_username(self, storage: SQLiteStorage) -> None:
        storage.add(_entry(username="alice@example.com"))
        storage.add(_entry(username="bob@example.com"))
        assert len(storage.list(search="alice")) == 1

    def test_search_by_url(self, storage: SQLiteStorage) -> None:
        storage.add(_entry(title="ServiceA", username="u", url="https://github.com"))
        storage.add(_entry(title="ServiceB", username="u", url="https://gitlab.com"))
        assert len(storage.list(search="github")) == 1

    def test_search_no_results(self, storage: SQLiteStorage) -> None:
        storage.add(_entry(title="GitHub"))
        assert storage.list(search="zzznomatch") == []

    def test_search_none_returns_all(self, storage: SQLiteStorage) -> None:
        storage.add(_entry(title="A"))
        storage.add(_entry(title="B"))
        assert len(storage.list(search=None)) == 2


# ---------------------------------------------------------------------------
# Update
# ---------------------------------------------------------------------------


class TestUpdate:
    def test_updates_title(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(title="Old"))
        entry.title = "New"
        storage.update(entry)
        assert storage.get(entry.id).title == "New"

    def test_updates_password(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(password="old-pw"))
        entry.password = "new-pw"
        storage.update(entry)
        assert storage.get(entry.id).password == "new-pw"

    def test_updates_email(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(email="old@example.com"))
        entry.email = "new@example.com"
        storage.update(entry)
        assert storage.get(entry.id).email == "new@example.com"

    def test_updates_notes(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(notes="old note"))
        entry.notes = "new note"
        storage.update(entry)
        assert storage.get(entry.id).notes == "new note"

    def test_updated_at_changes(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry())
        original_ts = entry.updated_at
        entry.title = "Changed"
        updated = storage.update(entry)
        assert updated.updated_at >= original_ts

    def test_raises_key_error_for_unknown_id(self, storage: SQLiteStorage) -> None:
        with pytest.raises(KeyError):
            storage.update(_entry(id=99999))

    def test_raises_value_error_for_none_id(self, storage: SQLiteStorage) -> None:
        with pytest.raises(ValueError, match="without id"):
            storage.update(_entry(id=None))


# ---------------------------------------------------------------------------
# Delete
# ---------------------------------------------------------------------------


class TestDelete:
    def test_deletes_entry(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry())
        storage.delete(entry.id)
        assert storage.get(entry.id) is None

    def test_raises_key_error_for_unknown_id(self, storage: SQLiteStorage) -> None:
        with pytest.raises(KeyError):
            storage.delete(99999)

    def test_other_entries_unaffected(self, storage: SQLiteStorage) -> None:
        e1 = storage.add(_entry(title="Keep"))
        e2 = storage.add(_entry(title="Delete"))
        storage.delete(e2.id)
        assert storage.get(e1.id) is not None
        assert len(storage.list()) == 1


# ---------------------------------------------------------------------------
# Company CRUD
# ---------------------------------------------------------------------------


class TestCompanyCRUD:
    def test_add_and_get(self, storage: SQLiteStorage) -> None:
        c = storage.add_company(_company(name="OpenAI"))
        assert c.id is not None
        fetched = storage.get_company(c.id)
        assert fetched.name == "OpenAI"

    def test_add_with_full_details(self, storage: SQLiteStorage) -> None:
        addr = CompanyAddress(country="United States", country_code="us",
                              street="1 Infinite Loop", city="Cupertino", state="CA",
                              postcode="95014")
        c = storage.add_company(Company(
            id=None, name="Apple Inc.",
            icon={"type": "iconify", "value": "logos:apple"},
            address=addr,
            revenue=383_000_000_000.0,
        ))
        fetched = storage.get_company(c.id)
        assert fetched.revenue == 383_000_000_000.0
        assert fetched.address.country == "United States"
        assert fetched.address.city == "Cupertino"
        assert fetched.icon["value"] == "logos:apple"

    def test_add_company_no_address(self, storage: SQLiteStorage) -> None:
        c = storage.add_company(_company(name="Bare Corp"))
        fetched = storage.get_company(c.id)
        assert fetched.address is None
        assert fetched.revenue is None
        assert fetched.icon is None

    def test_get_missing_returns_none(self, storage: SQLiteStorage) -> None:
        assert storage.get_company(99999) is None

    def test_list_companies_empty(self, storage: SQLiteStorage) -> None:
        assert storage.list_companies() == []

    def test_list_companies_ordered_by_name(self, storage: SQLiteStorage) -> None:
        storage.add_company(_company(name="Zebra"))
        storage.add_company(_company(name="Apple"))
        storage.add_company(_company(name="Meta"))
        names = [c.name for c in storage.list_companies()]
        assert names == sorted(names)

    def test_update_company(self, storage: SQLiteStorage) -> None:
        c = storage.add_company(_company(name="OldName"))
        c.name = "NewName"
        storage.update_company(c)
        assert storage.get_company(c.id).name == "NewName"

    def test_update_company_revenue(self, storage: SQLiteStorage) -> None:
        c = storage.add_company(_company(revenue=1_000_000.0))
        c.revenue = 2_000_000.0
        storage.update_company(c)
        assert storage.get_company(c.id).revenue == 2_000_000.0

    def test_update_company_raises_for_none_id(self, storage: SQLiteStorage) -> None:
        with pytest.raises(ValueError, match="without id"):
            storage.update_company(_company(id=None))

    def test_update_company_raises_for_unknown_id(self, storage: SQLiteStorage) -> None:
        with pytest.raises(KeyError):
            storage.update_company(_company(id=99999))

    def test_delete_company(self, storage: SQLiteStorage) -> None:
        c = storage.add_company(_company())
        storage.delete_company(c.id)
        assert storage.get_company(c.id) is None

    def test_delete_company_raises_for_unknown_id(self, storage: SQLiteStorage) -> None:
        with pytest.raises(KeyError):
            storage.delete_company(99999)

    def test_entry_company_id_set_null_on_delete(self, storage: SQLiteStorage) -> None:
        c = storage.add_company(_company(name="Parent"))
        e = storage.add(_entry(company_id=c.id))
        storage.delete_company(c.id)
        # ON DELETE SET NULL
        assert storage.get(e.id).company_id is None


# ---------------------------------------------------------------------------
# Icons
# ---------------------------------------------------------------------------


class TestIcons:
    def test_save_and_list(self, storage: SQLiteStorage) -> None:
        storage.save_icon("logo.svg", b"<svg/>")
        assert "logo.svg" in storage.list_icons()

    def test_icons_dir_property(self, storage: SQLiteStorage, tmp_path: Path) -> None:
        assert storage.icons_dir == tmp_path / "icons"


# ---------------------------------------------------------------------------
# Encryption
# ---------------------------------------------------------------------------


class TestEncryption:
    def test_wrong_master_fails_on_get(self, tmp_path: Path) -> None:
        s1 = SQLiteStorage(master_password="correct", vault_dir=tmp_path)
        s1.add(_entry(password="secret"))
        with pytest.raises(Exception):
            SQLiteStorage(master_password="wrong", vault_dir=tmp_path)

    def test_wrong_master_fails_on_list(self, tmp_path: Path) -> None:
        s1 = SQLiteStorage(master_password="correct", vault_dir=tmp_path)
        s1.add(_entry(password="secret"))
        with pytest.raises(Exception):
            SQLiteStorage(master_password="wrong", vault_dir=tmp_path)

    def test_correct_master_decrypts(self, tmp_path: Path) -> None:
        s1 = SQLiteStorage(master_password="correct", vault_dir=tmp_path)
        entry = s1.add(_entry(password="plaintext"))
        s2 = SQLiteStorage(master_password="correct", vault_dir=tmp_path)
        assert s2.get(entry.id).password == "plaintext"

    def test_to_dict_contains_all_keys(self, storage: SQLiteStorage) -> None:
        entry = storage.add(_entry(url="https://x.com", notes="n",
                                   email="x@x.com", category="Tech",
                                   tags=["a"], login_methods=["Google"]))
        d = entry.to_dict()
        for key in ("id", "title", "username", "email", "password", "url",
                    "notes", "category", "tags", "login_methods",
                    "created_at", "updated_at"):
            assert key in d


"""Tests for UIServer — server.py coverage."""

import time
from pathlib import Path

import pytest
import httpx

from sspwd.ui.server import UIServer


# ---------------------------------------------------------------------------
# App construction
# ---------------------------------------------------------------------------


class TestBuildApp:
    def test_app_created(self) -> None:
        server = UIServer(open_browser=False)
        assert server.app is not None

    def test_app_exposed_via_property(self) -> None:
        from fastapi import FastAPI
        server = UIServer(open_browser=False)
        assert isinstance(server.app, FastAPI)

    def test_static_dir_missing_no_crash(self) -> None:
        """Server builds fine even if ui/static/ has no index.html."""
        server = UIServer(open_browser=False)
        assert server.app is not None

    def test_static_dir_with_index_mounts(self, tmp_path: Path, monkeypatch) -> None:
        """Simulate a built frontend: fake static/index.html triggers the SPA mount."""
        from sspwd.ui import server as server_module
        from fastapi.testclient import TestClient

        fake_static = tmp_path / "fake_static"
        fake_static.mkdir()
        (fake_static / "index.html").write_text("<html>ok</html>")
        (fake_static / "assets").mkdir()

        monkeypatch.setattr(server_module, "_STATIC_DIR", fake_static)

        srv = UIServer(open_browser=False)
        client = TestClient(srv.app)

        r = client.get("/some/deep/route")
        assert r.status_code == 200
        assert "ok" in r.text


# ---------------------------------------------------------------------------
# start() non-blocking
# ---------------------------------------------------------------------------


class TestStart:
    def test_start_non_blocking_serves_health(self) -> None:
        port = 17523
        server = UIServer(host="127.0.0.1", port=port, open_browser=False)
        server.start(block=False)

        deadline = time.time() + 5.0
        last_exc: Exception = RuntimeError("server never started")
        while time.time() < deadline:
            try:
                r = httpx.get(f"http://127.0.0.1:{port}/api/v1/health", timeout=1.0)
                assert r.status_code == 200
                assert r.json()["status"] == "ok"
                return
            except Exception as exc:
                last_exc = exc
                time.sleep(0.15)
        raise last_exc

    def test_start_non_blocking_does_not_block_caller(self) -> None:
        port = 17524
        server = UIServer(host="127.0.0.1", port=port, open_browser=False)
        t0 = time.time()
        server.start(block=False)
        elapsed = time.time() - t0
        assert elapsed < 3.0, f"start(block=False) took {elapsed:.1f}s"

    def test_open_browser_false_does_not_call_webbrowser(self, monkeypatch) -> None:
        port = 17525
        called = []
        monkeypatch.setattr("webbrowser.open", lambda url: called.append(url))

        server = UIServer(host="127.0.0.1", port=port, open_browser=False)
        server.start(block=False)
        time.sleep(0.5)

        assert called == [], f"webbrowser.open was called with: {called}"


"""
SQLite storage backend — Argon2id + AES-256-GCM.

entries table columns:
    id, title, username*, email*, password*, url, notes*, category,
    tags (JSON), login_methods (JSON), company_id, user_created_at,
    created_at, updated_at
    (* encrypted with AES-256-GCM)

companies table columns:
    id, name, icon (JSON), address (JSON), revenue (REAL)
"""
from __future__ import annotations

import base64
import json
import os
import sqlite3
from datetime import datetime
from pathlib import Path
from typing import Optional

from argon2.low_level import Type, hash_secret_raw
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from .base import BaseStorage, Company, CompanyAddress, PasswordEntry

# ── constants ────────────────────────────────────────────────────────────────

_DB_FILENAME        = "vault.db"
_SALT_FILENAME      = "salt.bin"
_SENTINEL_FILENAME  = "verify.bin"
_SENTINEL_PLAINTEXT = b"sspwd-ok"
_ICONS_DIRNAME      = "icons"
_DEFAULT_PROJECT    = "default"
_NONCE_LEN          = 12

_ARGON2_TIME_COST   = 3
_ARGON2_MEMORY_COST = 65536
_ARGON2_PARALLELISM = 2
_ARGON2_HASH_LEN    = 32

# Columns that are AES-encrypted in the entries table
_ENCRYPTED_ENTRY_COLS = {"password", "email", "notes"}


# ── module helpers ────────────────────────────────────────────────────────────

def _derive_key(master_password: str, salt: bytes) -> bytes:
    return hash_secret_raw(
        secret=master_password.encode(),
        salt=salt,
        time_cost=_ARGON2_TIME_COST,
        memory_cost=_ARGON2_MEMORY_COST,
        parallelism=_ARGON2_PARALLELISM,
        hash_len=_ARGON2_HASH_LEN,
        type=Type.ID,
    )


def project_dir(project: str = _DEFAULT_PROJECT, base: Optional[Path] = None) -> Path:
    root = base or (Path.home() / ".sspwd")
    return root / project


# ── storage class ─────────────────────────────────────────────────────────────

class SQLiteStorage(BaseStorage):

    def __init__(
        self,
        master_password: str,
        project: str = _DEFAULT_PROJECT,
        vault_dir: Optional[Path] = None,
    ) -> None:
        self._vault_dir = Path(vault_dir) if vault_dir else project_dir(project)
        self._vault_dir.mkdir(parents=True, exist_ok=True)

        self._icons_dir     = self._vault_dir / _ICONS_DIRNAME
        self._icons_dir.mkdir(exist_ok=True)
        self._db_path       = self._vault_dir / _DB_FILENAME
        self._salt_path     = self._vault_dir / _SALT_FILENAME
        self._sentinel_path = self._vault_dir / _SENTINEL_FILENAME

        salt         = self._load_or_create_salt()
        key          = _derive_key(master_password, salt)
        self._aesgcm = AESGCM(key)

        self._write_or_verify_sentinel()
        self.initialize()

    # ── properties ───────────────────────────────────────────────────────────

    @property
    def icons_dir(self) -> Path:
        return self._icons_dir

    @property
    def vault_dir(self) -> Path:
        return self._vault_dir

    # ── crypto helpers ────────────────────────────────────────────────────────

    def _load_or_create_salt(self) -> bytes:
        if self._salt_path.exists():
            return self._salt_path.read_bytes()
        salt = os.urandom(32)
        self._salt_path.write_bytes(salt)
        return salt

    def _write_or_verify_sentinel(self) -> None:
        if self._sentinel_path.exists():
            raw = self._sentinel_path.read_bytes()
            self._aesgcm.decrypt(raw[:_NONCE_LEN], raw[_NONCE_LEN:], None)
        else:
            nonce = os.urandom(_NONCE_LEN)
            self._sentinel_path.write_bytes(
                nonce + self._aesgcm.encrypt(nonce, _SENTINEL_PLAINTEXT, None)
            )

    def _encrypt(self, plaintext: str) -> str:
        nonce = os.urandom(_NONCE_LEN)
        ct    = self._aesgcm.encrypt(nonce, plaintext.encode(), None)
        return base64.urlsafe_b64encode(nonce + ct).decode()

    def _decrypt(self, token: str) -> str:
        raw = base64.urlsafe_b64decode(token.encode())
        return self._aesgcm.decrypt(raw[:_NONCE_LEN], raw[_NONCE_LEN:], None).decode()

    # ── db helpers ────────────────────────────────────────────────────────────

    def _connect(self) -> sqlite3.Connection:
        conn = sqlite3.connect(self._db_path)
        conn.row_factory = sqlite3.Row
        conn.execute("PRAGMA foreign_keys = ON")
        return conn

    def _migrate(self, conn: sqlite3.Connection) -> None:
        """Add any columns that did not exist in older vault versions."""
        entry_existing  = {row[1] for row in conn.execute("PRAGMA table_info(entries)")}
        company_existing = {row[1] for row in conn.execute("PRAGMA table_info(companies)")}

        entry_new_cols = [
            ("email",           "TEXT"),
            ("category",        "TEXT DEFAULT 'Other'"),
            ("tags",            "TEXT DEFAULT '[]'"),
            ("login_methods",   "TEXT DEFAULT '[]'"),
            ("company_id",      "INTEGER"),
            ("user_created_at", "TEXT"),
        ]
        for col, defn in entry_new_cols:
            if col not in entry_existing:
                conn.execute(f"ALTER TABLE entries ADD COLUMN {col} {defn}")

        company_new_cols = [
            ("icon",    "TEXT"),
            ("address", "TEXT"),
            ("revenue", "REAL"),
        ]
        for col, defn in company_new_cols:
            if col not in company_existing:
                conn.execute(f"ALTER TABLE companies ADD COLUMN {col} {defn}")

    # ── schema ────────────────────────────────────────────────────────────────

    def initialize(self) -> None:
        with self._connect() as conn:
            conn.execute("""
                CREATE TABLE IF NOT EXISTS companies (
                    id      INTEGER PRIMARY KEY AUTOINCREMENT,
                    name    TEXT    NOT NULL,
                    icon    TEXT,           -- JSON {type, value}
                    address TEXT,           -- JSON CompanyAddress
                    revenue REAL            -- raw USD number
                )
            """)
            conn.execute("""
                CREATE TABLE IF NOT EXISTS entries (
                    id              INTEGER PRIMARY KEY AUTOINCREMENT,
                    title           TEXT    NOT NULL,
                    username        TEXT,
                    email           TEXT,   -- AES-encrypted
                    password        TEXT,   -- AES-encrypted
                    url             TEXT,
                    notes           TEXT,   -- AES-encrypted
                    category        TEXT    NOT NULL DEFAULT 'Other',
                    tags            TEXT    NOT NULL DEFAULT '[]',  -- JSON array
                    login_methods   TEXT    NOT NULL DEFAULT '[]',  -- JSON array
                    company_id      INTEGER REFERENCES companies(id) ON DELETE SET NULL,
                    user_created_at TEXT,
                    created_at      TEXT    NOT NULL,
                    updated_at      TEXT    NOT NULL
                )
            """)
            self._migrate(conn)

    # ── entries CRUD ──────────────────────────────────────────────────────────

    def _enc_opt(self, v: Optional[str]) -> Optional[str]:
        """Encrypt if not None, else return None."""
        return self._encrypt(v) if v is not None else None

    def _dec_opt(self, v: Optional[str]) -> Optional[str]:
        """Decrypt if not None, else return None."""
        return self._decrypt(v) if v is not None else None

    def _row_to_entry(self, row: sqlite3.Row) -> PasswordEntry:
        return PasswordEntry(
            id              = row["id"],
            title           = row["title"],
            username        = row["username"] if row["username"] else None,
            email           = self._dec_opt(row["email"]),
            password        = self._dec_opt(row["password"]),
            url             = row["url"],
            notes           = self._dec_opt(row["notes"]),
            category        = row["category"] or "Other",
            tags            = json.loads(row["tags"]) if row["tags"] else [],
            login_methods   = json.loads(row["login_methods"]) if row["login_methods"] else [],
            company_id      = row["company_id"],
            user_created_at = datetime.fromisoformat(row["user_created_at"]) if row["user_created_at"] else None,
            created_at      = datetime.fromisoformat(row["created_at"]),
            updated_at      = datetime.fromisoformat(row["updated_at"]),
        )

    def add(self, entry: PasswordEntry) -> PasswordEntry:
        now = datetime.utcnow().isoformat()
        with self._connect() as conn:
            cur = conn.execute(
                """INSERT INTO entries
                   (title, username, email, password, url, notes, category, tags,
                    login_methods, company_id, user_created_at, created_at, updated_at)
                   VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)""",
                (
                    entry.title,
                    entry.username,
                    self._enc_opt(entry.email),
                    self._enc_opt(entry.password),
                    entry.url,
                    self._enc_opt(entry.notes),
                    entry.category or "Other",
                    json.dumps(entry.tags),
                    json.dumps(entry.login_methods),
                    entry.company_id,
                    entry.user_created_at.isoformat() if entry.user_created_at else None,
                    now, now,
                ),
            )
            entry.id         = cur.lastrowid
            entry.created_at = datetime.fromisoformat(now)
            entry.updated_at = entry.created_at
        return entry

    def get(self, entry_id: int) -> Optional[PasswordEntry]:
        with self._connect() as conn:
            row = conn.execute("SELECT * FROM entries WHERE id=?", (entry_id,)).fetchone()
        return self._row_to_entry(row) if row else None

    def list(self, search: Optional[str] = None) -> list[PasswordEntry]:
        with self._connect() as conn:
            if search:
                p = f"%{search}%"
                rows = conn.execute(
                    """SELECT * FROM entries
                       WHERE title LIKE ? OR username LIKE ?
                          OR (url IS NOT NULL AND url LIKE ?)
                          OR category LIKE ?
                       ORDER BY title""",
                    (p, p, p, p),
                ).fetchall()
            else:
                rows = conn.execute("SELECT * FROM entries ORDER BY title").fetchall()
        return [self._row_to_entry(r) for r in rows]

    def update(self, entry: PasswordEntry) -> PasswordEntry:
        if entry.id is None:
            raise ValueError("Cannot update entry without id.")
        now = datetime.utcnow().isoformat()
        with self._connect() as conn:
            rc = conn.execute(
                """UPDATE entries
                   SET title=?, username=?, email=?, password=?, url=?, notes=?,
                       category=?, tags=?, login_methods=?, company_id=?,
                       user_created_at=?, updated_at=?
                   WHERE id=?""",
                (
                    entry.title,
                    entry.username,
                    self._enc_opt(entry.email),
                    self._enc_opt(entry.password),
                    entry.url,
                    self._enc_opt(entry.notes),
                    entry.category or "Other",
                    json.dumps(entry.tags),
                    json.dumps(entry.login_methods),
                    entry.company_id,
                    entry.user_created_at.isoformat() if entry.user_created_at else None,
                    now, entry.id,
                ),
            ).rowcount
        if rc == 0:
            raise KeyError(f"No entry with id={entry.id}")
        entry.updated_at = datetime.fromisoformat(now)
        return entry

    def delete(self, entry_id: int) -> None:
        with self._connect() as conn:
            rc = conn.execute("DELETE FROM entries WHERE id=?", (entry_id,)).rowcount
        if rc == 0:
            raise KeyError(f"No entry with id={entry_id}")

    # ── companies CRUD ────────────────────────────────────────────────────────

    def _row_to_company(self, row: sqlite3.Row) -> Company:
        icon_raw    = row["icon"]
        address_raw = row["address"]
        return Company(
            id      = row["id"],
            name    = row["name"],
            icon    = json.loads(icon_raw)    if icon_raw    else None,
            address = CompanyAddress.from_dict(json.loads(address_raw)) if address_raw else None,
            revenue = row["revenue"],
        )

    def add_company(self, company: Company) -> Company:
        with self._connect() as conn:
            cur = conn.execute(
                "INSERT INTO companies (name, icon, address, revenue) VALUES (?,?,?,?)",
                (
                    company.name,
                    json.dumps(company.icon)             if company.icon    else None,
                    json.dumps(company.address.to_dict()) if company.address else None,
                    company.revenue,
                ),
            )
            company.id = cur.lastrowid
        return company

    def get_company(self, company_id: int) -> Optional[Company]:
        with self._connect() as conn:
            row = conn.execute("SELECT * FROM companies WHERE id=?", (company_id,)).fetchone()
        return self._row_to_company(row) if row else None

    def list_companies(self) -> list[Company]:
        with self._connect() as conn:
            rows = conn.execute("SELECT * FROM companies ORDER BY name").fetchall()
        return [self._row_to_company(r) for r in rows]

    def update_company(self, company: Company) -> Company:
        if company.id is None:
            raise ValueError("Cannot update company without id.")
        with self._connect() as conn:
            rc = conn.execute(
                "UPDATE companies SET name=?, icon=?, address=?, revenue=? WHERE id=?",
                (
                    company.name,
                    json.dumps(company.icon)             if company.icon    else None,
                    json.dumps(company.address.to_dict()) if company.address else None,
                    company.revenue,
                    company.id,
                ),
            ).rowcount
        if rc == 0:
            raise KeyError(f"No company with id={company.id}")
        return company

    def delete_company(self, company_id: int) -> None:
        with self._connect() as conn:
            rc = conn.execute("DELETE FROM companies WHERE id=?", (company_id,)).rowcount
        if rc == 0:
            raise KeyError(f"No company with id={company_id}")

    # ── icons ─────────────────────────────────────────────────────────────────

    def save_icon(self, filename: str, data: bytes) -> Path:
        dest = self._icons_dir / filename
        dest.write_bytes(data)
        return dest

    def list_icons(self) -> list[str]:
        return [f.name for f in self._icons_dir.iterdir() if f.is_file()]


from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional


# ---------------------------------------------------------------------------
# Company / Owner
# ---------------------------------------------------------------------------

@dataclass
class CompanyAddress:
    country:      str
    country_code: str
    street:       Optional[str] = None
    city:         Optional[str] = None
    state:        Optional[str] = None
    postcode:     Optional[str] = None

    def to_dict(self) -> dict:
        return {
            "street":       self.street,
            "city":         self.city,
            "state":        self.state,
            "postcode":     self.postcode,
            "country":      self.country,
            "countryCode":  self.country_code,
        }

    @staticmethod
    def from_dict(d: dict) -> "CompanyAddress":
        return CompanyAddress(
            country      = d.get("country", ""),
            country_code = d.get("countryCode", ""),
            street       = d.get("street"),
            city         = d.get("city"),
            state        = d.get("state"),
            postcode     = d.get("postcode"),
        )


@dataclass
class Company:
    id:      Optional[int]
    name:    str
    icon:    Optional[dict] = None          # {type, value} — matches frontend IconSource
    address: Optional[CompanyAddress] = None
    revenue: Optional[float] = None         # raw USD number, e.g. 307_400_000_000.0

    def to_dict(self) -> dict:
        return {
            "id":      self.id,
            "name":    self.name,
            "icon":    self.icon,
            "address": self.address.to_dict() if self.address else None,
            "revenue": self.revenue,
        }

    @staticmethod
    def from_dict(d: dict) -> "Company":
        addr_raw = d.get("address")
        return Company(
            id      = d.get("id"),
            name    = d.get("name", ""),
            icon    = d.get("icon"),
            address = CompanyAddress.from_dict(addr_raw) if addr_raw else None,
            revenue = d.get("revenue"),
        )


# ---------------------------------------------------------------------------
# Password entry
# ---------------------------------------------------------------------------

@dataclass
class PasswordEntry:
    id:              Optional[int]
    title:           str
    username:        Optional[str]  = None   # login handle / display name
    email:           Optional[str]  = None   # login email
    password:        Optional[str]  = None   # stored encrypted, returned decrypted
    url:             Optional[str]  = None
    notes:           Optional[str]  = None
    category:        str            = "Other"
    tags:            list[str]      = field(default_factory=list)
    login_methods:   list[str]      = field(default_factory=list)
    company_id:      Optional[int]  = None
    user_created_at: Optional[datetime] = None
    created_at:      datetime       = field(default_factory=datetime.utcnow)
    updated_at:      datetime       = field(default_factory=datetime.utcnow)

    def to_dict(self) -> dict:
        return {
            "id":              self.id,
            "title":           self.title,
            "username":        self.username,
            "email":           self.email,
            "password":        self.password,
            "url":             self.url,
            "notes":           self.notes,
            "category":        self.category,
            "tags":            self.tags,
            "login_methods":   self.login_methods,
            "company_id":      self.company_id,
            "user_created_at": self.user_created_at.isoformat() if self.user_created_at else None,
            "created_at":      self.created_at.isoformat(),
            "updated_at":      self.updated_at.isoformat(),
        }


# ---------------------------------------------------------------------------
# Abstract storage interface
# ---------------------------------------------------------------------------

class BaseStorage(ABC):

    @abstractmethod
    def initialize(self) -> None: ...

    @abstractmethod
    def add(self, entry: PasswordEntry) -> PasswordEntry: ...

    @abstractmethod
    def get(self, entry_id: int) -> Optional[PasswordEntry]: ...

    @abstractmethod
    def list(self, search: Optional[str] = None) -> list[PasswordEntry]: ...

    @abstractmethod
    def update(self, entry: PasswordEntry) -> PasswordEntry: ...

    @abstractmethod
    def delete(self, entry_id: int) -> None: ...

    def add_company(self, company: Company) -> Company:
        raise NotImplementedError

    def get_company(self, company_id: int) -> Optional[Company]:
        raise NotImplementedError

    def list_companies(self) -> list[Company]:
        raise NotImplementedError

    def update_company(self, company: Company) -> Company:
        raise NotImplementedError

    def delete_company(self, company_id: int) -> None:
        raise NotImplementedError


"""Integration tests for the REST API — targets 100% api.py coverage."""

import pytest
from pathlib import Path
from fastapi.testclient import TestClient

from sspwd.storage.sqlite import SQLiteStorage
from sspwd.storage.base import Company, CompanyAddress
import sspwd.ui.api as api_module
from sspwd.ui.server import UIServer

PROJECT = "testproject"


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def client(tmp_path: Path) -> TestClient:
    """Build a TestClient with one project pre-unlocked in _sessions."""
    storage = SQLiteStorage(master_password="test", vault_dir=tmp_path)
    api_module._sessions[PROJECT] = storage
    server = UIServer(open_browser=False)
    c = TestClient(server.app)
    yield c
    api_module._sessions.pop(PROJECT, None)


@pytest.fixture
def client_with_entry(client: TestClient) -> tuple[TestClient, dict]:
    """Client that already has one entry created."""
    payload = {"title": "GitHub", "username": "alice", "password": "s3cr3t",
               "url": "https://github.com"}
    entry = client.post(f"/api/v1/entries?project={PROJECT}", json=payload).json()
    return client, entry


@pytest.fixture
def client_with_company(client: TestClient) -> tuple[TestClient, dict]:
    """Client that already has one company created."""
    payload = {
        "name": "Acme Corp",
        "icon": {"type": "letter", "value": "A"},
        "address": {
            "street": "1 Main St", "city": "Springfield",
            "state": "IL", "postcode": "62701",
            "country": "United States", "countryCode": "us",
        },
        "revenue": 1_000_000.0,
    }
    company = client.post(f"/api/v1/companies?project={PROJECT}", json=payload).json()
    return client, company


# ---------------------------------------------------------------------------
# Health
# ---------------------------------------------------------------------------


def test_health(client: TestClient) -> None:
    r = client.get("/api/v1/health")
    assert r.status_code == 200
    assert r.json() == {"status": "ok"}


# ---------------------------------------------------------------------------
# set_storage helper (covers api.py line 30)
# ---------------------------------------------------------------------------


def test_set_storage(tmp_path: Path) -> None:
    storage = SQLiteStorage(master_password="pw", vault_dir=tmp_path)
    api_module.set_storage(storage)
    assert api_module._sessions["default"] is storage
    api_module._sessions.pop("default", None)


# ---------------------------------------------------------------------------
# GET /entries
# ---------------------------------------------------------------------------


def test_list_empty(client: TestClient) -> None:
    r = client.get(f"/api/v1/entries?project={PROJECT}")
    assert r.status_code == 200
    assert r.json() == []


def test_list_returns_all(client: TestClient) -> None:
    for title in ("Alpha", "Beta", "Gamma"):
        client.post(f"/api/v1/entries?project={PROJECT}",
                    json={"title": title, "username": "u", "password": "p"})
    r = client.get(f"/api/v1/entries?project={PROJECT}")
    assert len(r.json()) == 3


def test_list_search_by_title(client: TestClient) -> None:
    for title in ("GitHub", "GitLab", "AWS"):
        client.post(f"/api/v1/entries?project={PROJECT}",
                    json={"title": title, "username": "u", "password": "p"})
    r = client.get(f"/api/v1/entries?project={PROJECT}&search=Git")
    assert r.status_code == 200
    assert len(r.json()) == 2


def test_list_search_by_username(client: TestClient) -> None:
    client.post(f"/api/v1/entries?project={PROJECT}",
                json={"title": "A", "username": "alice@example.com", "password": "p"})
    client.post(f"/api/v1/entries?project={PROJECT}",
                json={"title": "B", "username": "bob@example.com", "password": "p"})
    r = client.get(f"/api/v1/entries?project={PROJECT}&search=alice")
    assert len(r.json()) == 1
    assert r.json()[0]["username"] == "alice@example.com"


def test_list_search_no_results(client: TestClient) -> None:
    client.post(f"/api/v1/entries?project={PROJECT}",
                json={"title": "GitHub", "username": "u", "password": "p"})
    r = client.get(f"/api/v1/entries?project={PROJECT}&search=zzznomatch")
    assert r.json() == []


# ---------------------------------------------------------------------------
# POST /entries
# ---------------------------------------------------------------------------


def test_create_title_only(client: TestClient) -> None:
    """Only title is required — all other fields are optional."""
    r = client.post(f"/api/v1/entries?project={PROJECT}", json={"title": "Bare"})
    assert r.status_code == 201
    data = r.json()
    assert data["title"] == "Bare"
    assert data["username"] is None
    assert data["password"] is None


def test_create_missing_title_returns_422(client: TestClient) -> None:
    """Omitting title (the only required field) must return 422."""
    r = client.post(f"/api/v1/entries?project={PROJECT}", json={})
    assert r.status_code == 422


def test_create_full(client: TestClient) -> None:
    payload = {
        "title": "AWS", "username": "admin", "email": "admin@example.com",
        "password": "hunter2", "url": "https://aws.amazon.com",
        "notes": "root account", "category": "Cloud",
        "tags": ["work", "critical"],
        "login_methods": ["Email / Password"],
        "user_created_at": "2020-01-01T00:00:00",
    }
    r = client.post(f"/api/v1/entries?project={PROJECT}", json=payload)
    assert r.status_code == 201
    data = r.json()
    assert data["url"] == "https://aws.amazon.com"
    assert data["notes"] == "root account"
    assert data["email"] == "admin@example.com"
    assert data["category"] == "Cloud"
    assert data["tags"] == ["work", "critical"]
    assert data["login_methods"] == ["Email / Password"]


def test_create_timestamps_present(client: TestClient) -> None:
    r = client.post(f"/api/v1/entries?project={PROJECT}",
                    json={"title": "T", "username": "u", "password": "p"})
    data = r.json()
    assert "created_at" in data
    assert "updated_at" in data


def test_create_with_invalid_user_created_at(client: TestClient) -> None:
    """Malformed user_created_at is silently ignored (parsed as None)."""
    r = client.post(f"/api/v1/entries?project={PROJECT}",
                    json={"title": "T", "user_created_at": "not-a-date"})
    assert r.status_code == 201
    assert r.json()["user_created_at"] is None


# ---------------------------------------------------------------------------
# GET /entries/{id}
# ---------------------------------------------------------------------------


def test_get_entry(client_with_entry) -> None:
    client, entry = client_with_entry
    r = client.get(f"/api/v1/entries/{entry['id']}?project={PROJECT}")
    assert r.status_code == 200
    assert r.json()["username"] == "alice"


def test_get_entry_not_found(client: TestClient) -> None:
    r = client.get(f"/api/v1/entries/99999?project={PROJECT}")
    assert r.status_code == 404
    assert "not found" in r.json()["detail"].lower()


# ---------------------------------------------------------------------------
# PUT /entries/{id}
# ---------------------------------------------------------------------------


def test_update_entry(client_with_entry) -> None:
    client, entry = client_with_entry
    r = client.put(f"/api/v1/entries/{entry['id']}?project={PROJECT}",
                   json={"title": "GitHub Pro", "username": "alice2", "password": "newpass"})
    assert r.status_code == 200
    data = r.json()
    assert data["title"] == "GitHub Pro"
    assert data["username"] == "alice2"


def test_update_entry_not_found(client: TestClient) -> None:
    r = client.put(f"/api/v1/entries/99999?project={PROJECT}",
                   json={"title": "X", "username": "x", "password": "x"})
    assert r.status_code == 404
    assert "not found" in r.json()["detail"].lower()


def test_update_persists(client_with_entry) -> None:
    client, entry = client_with_entry
    client.put(f"/api/v1/entries/{entry['id']}?project={PROJECT}",
               json={"title": "Updated", "username": "u", "password": "p"})
    fetched = client.get(f"/api/v1/entries/{entry['id']}?project={PROJECT}").json()
    assert fetched["title"] == "Updated"


# ---------------------------------------------------------------------------
# DELETE /entries/{id}
# ---------------------------------------------------------------------------


def test_delete_entry(client_with_entry) -> None:
    client, entry = client_with_entry
    r = client.delete(f"/api/v1/entries/{entry['id']}?project={PROJECT}")
    assert r.status_code == 204
    r2 = client.get(f"/api/v1/entries/{entry['id']}?project={PROJECT}")
    assert r2.status_code == 404


def test_delete_entry_not_found(client: TestClient) -> None:
    r = client.delete(f"/api/v1/entries/99999?project={PROJECT}")
    assert r.status_code == 404
    assert "not found" in r.json()["detail"].lower()


def test_delete_reduces_list(client: TestClient) -> None:
    ids = [
        client.post(f"/api/v1/entries?project={PROJECT}",
                    json={"title": f"E{i}", "username": "u", "password": "p"}).json()["id"]
        for i in range(3)
    ]
    client.delete(f"/api/v1/entries/{ids[0]}?project={PROJECT}")
    assert len(client.get(f"/api/v1/entries?project={PROJECT}").json()) == 2


# ---------------------------------------------------------------------------
# Companies CRUD
# ---------------------------------------------------------------------------


def test_list_companies_empty(client: TestClient) -> None:
    r = client.get(f"/api/v1/companies?project={PROJECT}")
    assert r.status_code == 200
    assert r.json() == []


def test_create_company_minimal(client: TestClient) -> None:
    r = client.post(f"/api/v1/companies?project={PROJECT}",
                    json={"name": "Bare Corp"})
    assert r.status_code == 201
    data = r.json()
    assert data["name"] == "Bare Corp"
    assert data["id"] is not None


def test_create_company_full(client_with_company) -> None:
    client, company = client_with_company
    assert company["name"] == "Acme Corp"
    assert company["address"]["city"] == "Springfield"
    assert company["revenue"] == 1_000_000.0
    assert company["icon"]["type"] == "letter"


def test_get_company(client_with_company) -> None:
    client, company = client_with_company
    r = client.get(f"/api/v1/companies/{company['id']}?project={PROJECT}")
    assert r.status_code == 200
    assert r.json()["name"] == "Acme Corp"


def test_get_company_not_found(client: TestClient) -> None:
    r = client.get(f"/api/v1/companies/99999?project={PROJECT}")
    assert r.status_code == 404


def test_update_company(client_with_company) -> None:
    client, company = client_with_company
    r = client.put(f"/api/v1/companies/{company['id']}?project={PROJECT}",
                   json={"name": "Acme Global", "revenue": 5_000_000.0})
    assert r.status_code == 200
    assert r.json()["name"] == "Acme Global"


def test_update_company_not_found(client: TestClient) -> None:
    r = client.put(f"/api/v1/companies/99999?project={PROJECT}",
                   json={"name": "Ghost"})
    assert r.status_code == 404


def test_delete_company(client_with_company) -> None:
    client, company = client_with_company
    r = client.delete(f"/api/v1/companies/{company['id']}?project={PROJECT}")
    assert r.status_code == 204
    assert client.get(f"/api/v1/companies/{company['id']}?project={PROJECT}").status_code == 404


def test_delete_company_not_found(client: TestClient) -> None:
    r = client.delete(f"/api/v1/companies/99999?project={PROJECT}")
    assert r.status_code == 404


def test_create_company_no_address_country(client: TestClient) -> None:
    """address dict without country → address stored as None (covers _company_in_to_obj branch)."""
    r = client.post(f"/api/v1/companies?project={PROJECT}",
                    json={"name": "NoCountry", "address": {"street": "123 Main"}})
    assert r.status_code == 201
    assert r.json()["address"] is None


# ---------------------------------------------------------------------------
# Session / lock guard
# ---------------------------------------------------------------------------


def test_require_raises_when_project_locked() -> None:
    from fastapi import HTTPException
    with pytest.raises(HTTPException) as exc_info:
        api_module._require("__nonexistent_project__")
    assert exc_info.value.status_code == 401


def test_unlock_wrong_password(tmp_path: Path) -> None:
    vault_path = tmp_path / "myproject"
    SQLiteStorage(master_password="correct", vault_dir=vault_path)
    server = UIServer(open_browser=False)
    client = TestClient(server.app)
    from unittest.mock import patch
    with patch("sspwd.ui.api.project_dir", return_value=vault_path), \
         patch("sspwd.storage.sqlite.project_dir", return_value=vault_path):
        r = client.post("/api/v1/projects/myproject/unlock",
                        json={"password": "wrongpassword"})
    assert r.status_code == 401


def test_unlock_project_not_found(tmp_path: Path) -> None:
    server = UIServer(open_browser=False)
    client = TestClient(server.app)
    r = client.post("/api/v1/projects/__ghost__/unlock", json={"password": "pw"})
    assert r.status_code == 404


def test_create_project_and_list(tmp_path: Path) -> None:
    server = UIServer(open_browser=False)
    client = TestClient(server.app)
    from unittest.mock import patch
    fake_dir = tmp_path / "newproj"
    with patch("sspwd.ui.api.project_dir", return_value=fake_dir), \
         patch("sspwd.ui.api.SQLiteStorage") as MockStorage:
        instance = MockStorage.return_value
        instance.list.return_value = []
        r = client.post("/api/v1/projects", json={"name": "newproj", "password": "pw"})
    assert r.status_code == 201
    assert r.json()["project"] == "newproj"


def test_create_project_invalid_name() -> None:
    server = UIServer(open_browser=False)
    client = TestClient(server.app)
    r = client.post("/api/v1/projects", json={"name": "bad/name", "password": "pw"})
    assert r.status_code == 400


def test_create_project_already_exists(tmp_path: Path) -> None:
    vault_path = tmp_path / "exists"
    SQLiteStorage(master_password="pw", vault_dir=vault_path)
    server = UIServer(open_browser=False)
    client = TestClient(server.app)
    from unittest.mock import patch
    with patch("sspwd.ui.api.project_dir", return_value=vault_path):
        r = client.post("/api/v1/projects", json={"name": "exists", "password": "pw"})
    assert r.status_code == 409


def test_list_unlocked(client: TestClient) -> None:
    r = client.get("/api/v1/projects/unlocked")
    assert r.status_code == 200
    assert PROJECT in r.json()


def test_lock_project(client: TestClient) -> None:
    api_module._sessions["to_lock"] = api_module._sessions[PROJECT]
    r = client.delete("/api/v1/projects/to_lock/lock")
    assert r.status_code == 200
    assert "to_lock" not in api_module._sessions


def test_unlock_success(tmp_path: Path) -> None:
    vault_path = tmp_path / "goodproject"
    SQLiteStorage(master_password="correct", vault_dir=vault_path)
    server = UIServer(open_browser=False)
    client = TestClient(server.app)
    from unittest.mock import patch
    with patch("sspwd.ui.api.project_dir", return_value=vault_path), \
         patch("sspwd.storage.sqlite.project_dir", return_value=vault_path):
        r = client.post("/api/v1/projects/goodproject/unlock",
                        json={"password": "correct"})
    assert r.status_code == 200
    assert r.json()["status"] == "unlocked"
    api_module._sessions.pop("goodproject", None)


def test_list_projects_no_sspwd_dir(tmp_path: Path, monkeypatch) -> None:
    import pathlib
    empty_home = tmp_path / "empty_home"
    empty_home.mkdir()
    monkeypatch.setattr(pathlib.Path, "home", classmethod(lambda cls: empty_home))
    server = UIServer(open_browser=False)
    client = TestClient(server.app)
    r = client.get("/api/v1/projects")
    assert r.status_code == 200
    assert r.json() == []


# ---------------------------------------------------------------------------
# Icons
# ---------------------------------------------------------------------------


def test_icon_upload_and_serve(client: TestClient) -> None:
    png_bytes = (
        b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01"
        b"\x00\x00\x00\x01\x08\x02\x00\x00\x00\x90wS\xde\x00\x00"
        b"\x00\x0cIDATx\x9cc\xf8\x0f\x00\x00\x01\x01\x00\x05\x18"
        b"\xd8N\x00\x00\x00\x00IEND\xaeB`\x82"
    )
    r = client.post(
        f"/api/v1/icons?project={PROJECT}",
        files={"file": ("test.png", png_bytes, "image/png")},
    )
    assert r.status_code == 201
    icon_url = r.json()["url"]
    assert icon_url.startswith("/api/v1/icons/")
    r2 = client.get(icon_url)
    assert r2.status_code == 200


def test_icon_upload_jpeg_extension(client: TestClient) -> None:
    """JPEG uploads get a .jpg extension (covers the .jpe/.jfif normalisation branch)."""
    # Minimal JPEG header
    jpeg_bytes = bytes([0xFF, 0xD8, 0xFF, 0xE0]) + b"\x00" * 16
    r = client.post(
        f"/api/v1/icons?project={PROJECT}",
        files={"file": ("photo.jpg", jpeg_bytes, "image/jpeg")},
    )
    assert r.status_code == 201
    assert r.json()["filename"].endswith(".jpg")


def test_icon_upload_unsupported_type(client: TestClient) -> None:
    r = client.post(
        f"/api/v1/icons?project={PROJECT}",
        files={"file": ("bad.txt", b"hello", "text/plain")},
    )
    assert r.status_code == 415


def test_icon_upload_too_large(client: TestClient) -> None:
    big = b"x" * (2 * 1024 * 1024 + 1)
    r = client.post(
        f"/api/v1/icons?project={PROJECT}",
        files={"file": ("big.png", big, "image/png")},
    )
    assert r.status_code == 413


def test_icon_serve_not_found(client: TestClient) -> None:
    r = client.get(f"/api/v1/icons/nonexistent.png?project={PROJECT}")
    assert r.status_code == 404


def test_icon_list(client: TestClient) -> None:
    r = client.get(f"/api/v1/icons?project={PROJECT}")
    assert r.status_code == 200
    assert isinstance(r.json(), list)


def test_icon_list_after_upload(client: TestClient) -> None:
    """GET /icons returns the uploaded file after a successful upload."""
    png = b"\x89PNG\r\n\x1a\n" + b"\x00" * 20
    client.post(f"/api/v1/icons?project={PROJECT}",
                files={"file": ("x.png", png, "image/png")})
    r = client.get(f"/api/v1/icons?project={PROJECT}")
    assert len(r.json()) >= 1


"""CLI tests using Click's CliRunner — targets cli.py from 0% → ~85% coverage.

`serve` is excluded because it starts a blocking uvicorn server;
that path is covered indirectly by test_api.py via UIServer.
"""

import pytest
from pathlib import Path
from click.testing import CliRunner

from sspwd.cli import cli
from sspwd.storage.sqlite import SQLiteStorage
from sspwd.storage.base import PasswordEntry


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _seed(vault_dir: Path, master: str = "pw") -> PasswordEntry:
    """Create a vault with a single entry and return it."""
    storage = SQLiteStorage(master_password=master, vault_dir=vault_dir)
    return storage.add(PasswordEntry(
        id=None, title="GitHub", username="alice", password="s3cr3t",
        url="https://github.com", notes="work account",
    ))


# ---------------------------------------------------------------------------
# version
# ---------------------------------------------------------------------------


def test_version() -> None:
    runner = CliRunner()
    result = runner.invoke(cli, ["version"])
    assert result.exit_code == 0
    assert "sspwd" in result.output
    assert result.output.strip() != ""  # version string present, not pinned


def test_version_flag() -> None:
    runner = CliRunner()
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    # just check it contains a semver-like pattern, not a hardcoded string
    import re
    assert re.search(r"\d+\.\d+\.\d+", result.output)


# ---------------------------------------------------------------------------
# list
# ---------------------------------------------------------------------------


class TestListCommand:
    def test_empty_vault(self, tmp_path: Path) -> None:
        # initialise the vault so the salt exists, but add no entries
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        runner = CliRunner()
        result = runner.invoke(cli, ["list", "--vault-dir", str(tmp_path)], input="pw\n")
        assert result.exit_code == 0
        assert "No entries found" in result.output

    def test_lists_entry(self, tmp_path: Path) -> None:
        _seed(tmp_path)
        runner = CliRunner()
        result = runner.invoke(cli, ["list", "--vault-dir", str(tmp_path)], input="pw\n")
        assert result.exit_code == 0
        assert "GitHub" in result.output
        assert "alice" in result.output

    def test_search_flag_match(self, tmp_path: Path) -> None:
        _seed(tmp_path)
        runner = CliRunner()
        result = runner.invoke(
            cli, ["list", "--vault-dir", str(tmp_path), "--search", "GitHub"], input="pw\n"
        )
        assert "GitHub" in result.output

    def test_search_flag_no_match(self, tmp_path: Path) -> None:
        _seed(tmp_path)
        runner = CliRunner()
        result = runner.invoke(
            cli, ["list", "--vault-dir", str(tmp_path), "--search", "zzznomatch"], input="pw\n"
        )
        assert "No entries found" in result.output


# ---------------------------------------------------------------------------
# add
# ---------------------------------------------------------------------------


class TestAddCommand:
    def test_add_entry(self, tmp_path: Path) -> None:
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)  # init salt
        runner = CliRunner()
        # prompt order: master pw, title, category, username, password (×2), url, notes
        inputs = "pw\nNotion\nSoftware\nbob@example.com\nmypassword\nmypassword\nhttps://notion.so\n\n"
        result = runner.invoke(cli, ["add", "--vault-dir", str(tmp_path)], input=inputs)
        assert result.exit_code == 0
        assert "Saved" in result.output

    def test_add_persists_to_storage(self, tmp_path: Path) -> None:
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        runner = CliRunner()
        # prompt order: master pw, title, category, username, password (×2), url, notes
        inputs = "pw\nFigma\nDesign\ncarol@example.com\nsecretpw\nsecretpw\n\n\n"
        runner.invoke(cli, ["add", "--vault-dir", str(tmp_path)], input=inputs)

        storage = SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        entries = storage.list()
        assert any(e.title == "Figma" for e in entries)


# ---------------------------------------------------------------------------
# get
# ---------------------------------------------------------------------------


class TestGetCommand:
    def test_get_existing_entry(self, tmp_path: Path) -> None:
        entry = _seed(tmp_path)
        runner = CliRunner()
        result = runner.invoke(cli, ["get", str(entry.id), "--vault-dir", str(tmp_path)], input="pw\n")
        assert result.exit_code == 0
        assert "GitHub" in result.output
        assert "alice" in result.output
        assert "s3cr3t" in result.output
        assert "https://github.com" in result.output
        assert "work account" in result.output

    def test_get_not_found_exits_1(self, tmp_path: Path) -> None:
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        runner = CliRunner()
        result = runner.invoke(cli, ["get", "99999", "--vault-dir", str(tmp_path)], input="pw\n")
        assert result.exit_code == 1
        assert "not found" in result.output.lower()


# ---------------------------------------------------------------------------
# delete
# ---------------------------------------------------------------------------


class TestDeleteCommand:
    def test_delete_with_yes_flag(self, tmp_path: Path) -> None:
        entry = _seed(tmp_path)
        runner = CliRunner()
        result = runner.invoke(
            cli, ["delete", str(entry.id), "--yes", "--vault-dir", str(tmp_path)], input="pw\n"
        )
        assert result.exit_code == 0
        assert "Deleted" in result.output

    def test_delete_with_confirmation(self, tmp_path: Path) -> None:
        entry = _seed(tmp_path)
        runner = CliRunner()
        # inputs: master password, then "y" to confirm
        result = runner.invoke(
            cli, ["delete", str(entry.id), "--vault-dir", str(tmp_path)], input="pw\ny\n"
        )
        assert result.exit_code == 0
        assert "Deleted" in result.output

    def test_delete_abort_on_no(self, tmp_path: Path) -> None:
        entry = _seed(tmp_path)
        runner = CliRunner()
        result = runner.invoke(
            cli, ["delete", str(entry.id), "--vault-dir", str(tmp_path)], input="pw\nn\n"
        )
        # Click raises Abort on "n", which exits non-zero
        assert result.exit_code != 0
        # Entry must still exist
        storage = SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        assert storage.get(entry.id) is not None

    def test_delete_not_found_exits_1(self, tmp_path: Path) -> None:
        SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        runner = CliRunner()
        result = runner.invoke(
            cli, ["delete", "99999", "--yes", "--vault-dir", str(tmp_path)], input="pw\n"
        )
        assert result.exit_code == 1
        assert "not found" in result.output.lower()

    def test_delete_removes_entry(self, tmp_path: Path) -> None:
        entry = _seed(tmp_path)
        runner = CliRunner()
        runner.invoke(
            cli, ["delete", str(entry.id), "--yes", "--vault-dir", str(tmp_path)], input="pw\n"
        )
        storage = SQLiteStorage(master_password="pw", vault_dir=tmp_path)
        assert storage.get(entry.id) is None


"""
UIServer — no master password at startup.
Projects are unlocked on demand via the API.
"""

from __future__ import annotations

import threading
import webbrowser
from pathlib import Path

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from fastapi.responses import FileResponse

from .api import router as api_router

_STATIC_DIR = Path(__file__).parent / "static"


class UIServer:
    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 7523,
        open_browser: bool = True,
    ) -> None:
        self._host = host
        self._port = port
        self._open_browser = open_browser
        self._app = self._build_app()

    def _build_app(self) -> FastAPI:
        app = FastAPI(title="sspwd", docs_url="/api/docs", redoc_url=None)

        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_methods=["*"],
            allow_headers=["*"],
        )

        app.include_router(api_router)

        if _STATIC_DIR.is_dir() and (_STATIC_DIR / "index.html").exists():
            if (_STATIC_DIR / "assets").is_dir():
                app.mount("/assets", StaticFiles(directory=_STATIC_DIR / "assets"), name="assets")

            @app.get("/{full_path:path}", include_in_schema=False)
            async def spa_fallback(full_path: str):
                return FileResponse(_STATIC_DIR / "index.html")

        return app

    def start(self, block: bool = True) -> None:
        url = f"http://{self._host}:{self._port}"

        config = uvicorn.Config(self._app, host=self._host, port=self._port, log_level="warning")
        server = uvicorn.Server(config)

        if self._open_browser:
            original_startup = server.startup

            async def _startup_and_open(sockets=None):
                await original_startup(sockets=sockets)
                threading.Thread(target=webbrowser.open, args=(url,), daemon=True).start()

            server.startup = _startup_and_open  # type: ignore[method-assign]

        print(f"sspwd UI → {url}  (Ctrl+C to quit)")

        if block:
            server.run()
        else:
            threading.Thread(target=server.run, daemon=True).start()

    @property
    def app(self) -> FastAPI:
        return self._app